from typing import Any, Iterable, List

from fugue_sql_antlr import FugueSQLParser
from pytest import raises
from triad import to_uuid
from triad.collections.schema import Schema
from triad.utils.convert import get_caller_global_local_vars

from fugue import (
    DataFrames,
    FugueWorkflow,
    LocalDataFrame,
    OutputTransformer,
    PartitionSpec,
    WorkflowDataFrame,
    WorkflowDataFrames,
    module,
    register_sql_engine,
)
from fugue.extensions.transformer.convert import _to_output_transformer
from fugue.sql._visitors import FugueSQLHooks, _Extensions
from fugue.exceptions import FugueSQLError
from fugue.execution.native_execution_engine import QPDPandasEngine


def test_create_data():
    w = FugueWorkflow().df([[0], [1]], "a:int", data_determiner=to_uuid)
    assert_eq(
        """
    a=create [[0],[1]] schema a:int
    """,
        w.workflow,
    )


def test_create():
    dag = FugueWorkflow()
    dag.create(mock_create1, params=dict(n=1))
    dag.create(mock_create2, schema="a:int", params=dict(n=1))
    assert_eq(
        """
    a=create using mock_create1 params n:1
    b=create using mock_create2(n=1) schema a:int
    """,
        dag,
    )


def test_process():
    # basic features, nest
    dag = FugueWorkflow()
    a1 = dag.create(mock_create1, params=dict(n=1))
    a2 = dag.create(mock_create1, params=dict(n=2))
    dag.process(a1, a2, using=mock_processor1, params=dict(n=3))
    dag.process(a2, a1, using=mock_processor2, schema="b:int", params=dict(n=4))
    dag.process(
        dag.create(mock_create1, params=dict(n=5)),
        dag.create(mock_create1, params=dict(n=6)),
        using=mock_processor1,
        params=dict(n=7),
    )
    assert_eq(
        """
    a=create using mock_create1 params n:1
    b=create using mock_create1 params n:2
    process a,b using mock_processor1(n=3)
    process b,a using mock_processor2(n=4) schema b:int
    process  # nested
        (create using mock_create1(n=5)),
        (create using mock_create1(n=6))
        using mock_processor1(n=7)
    """,
        dag,
    )

    # anonymous, nested anonymous
    dag = FugueWorkflow()
    a = dag.create(mock_create1, params=dict(n=1)).process(mock_processor3)
    b = a.partition(by=["a"]).process(mock_processor3)
    c = a.process(mock_processor3)
    dag.process(b, c, using=mock_processor1)
    assert_eq(
        """
    create using mock_create1 params n:1
    process using mock_processor3
    process  # nested
        (process prepartition by a using mock_processor3),
        (process using mock_processor3)
        using mock_processor1
    """,
        dag,
    )

    # no last dataframe
    with raises(FugueSQLError):
        assert_eq(
            """
        process using mock_processor3
        """,
            None,
        )

    # dict like dataframes
    dag = FugueWorkflow()
    a = dag.create(mock_create1, params=dict(n=1))
    b = dag.create(mock_create1, params=dict(n=2))
    dag.process(dict(df1=a, df2=b), using=mock_processor1)
    assert_eq(
        """
    process
        df1=(create using mock_create1(n=1)),
        df2:(create using mock_create1(n=2))
        using mock_processor1
    """,
        dag,
    )


def test_zip():
    dag = FugueWorkflow()
    a1 = dag.create(mock_create1, params=dict(n=1))
    a2 = dag.create(mock_create1, params=dict(n=2))
    a1.zip(a2)
    assert_eq(
        """
    a=create using mock_create1 params n:1
    zip a,(create using mock_create1 params n:2)
    """,
        dag,
    )

    dag = FugueWorkflow()
    a1 = dag.create(mock_create1, params=dict(n=1))
    a2 = dag.create(mock_create1, params=dict(n=2))
    a1.zip(a2, how="left_outer", partition=dict(by=["a"], presort="b DESC"))
    assert_eq(
        """
    a=create using mock_create1 params n:1
    zip a,(create using mock_create1 params n:2) left
        outer by a presort b desc
    """,
        dag,
    )


def test_cotransform():
    dag = FugueWorkflow()
    a1 = dag.create(mock_create1, params=dict(n=1))
    a2 = dag.create(mock_create1, params=dict(n=2))
    z = dag.zip(a1, a2)
    t = z.partition(num=3).transform(mock_cotransformer1, params=dict(n=3))
    assert_eq(
        """
    zip
        (create using mock_create1 params n:1),
        (create using mock_create1 params n:2)
    transform prepartition 3 using mock_cotransformer1(n=3)
    """,
        dag,
    )


def test_transform():
    w = (
        FugueWorkflow()
        .df([[0], [1]], "a:int", data_determiner=to_uuid)
        .transform(mock_transformer, schema=Schema("a:int"), params=dict(n=2))
    )
    assert_eq(
        """
    create [[0],[1]] schema a:int
    transform using mock_transformer(n=2) schema a:int
    """,
        w.workflow,
    )

    w = (
        FugueWorkflow()
        .df([[0], [1]], "a:int", data_determiner=to_uuid)
        .partition(by=["a"], presort="b DESC", num="ROWCOUNT/2")
        .transform(mock_transformer, schema="*~k+AA:str,bb:int", params=dict(n=2))
    )
    assert_eq(
        """
    create [[0],[1]] schema a:int

    transform
        prepartition ROWCOUNT / 2 by a presort b desc
        using mock_transformer(n=2) schema * ~ k + AA : str,bb:int
    """,
        w.workflow,
    )

    def _func(a: int, b: int) -> int:
        return a + b

    w = (
        FugueWorkflow()
        .df([[0], [1]], "a:int", data_determiner=to_uuid)
        .partition(by=["a"], presort="b DESC", num="ROWCOUNT/2")
        .transform(mock_transformer, schema="*", params=dict(n=2), callback=_func)
    )
    assert_eq(
        """
    create [[0],[1]] schema a:int

    transform
        prepartition ROWCOUNT / 2 by a presort b desc
        using mock_transformer(n=2) schema *
        callback _func
    """,
        w.workflow,
    )


def test_out_transform():
    class OT(OutputTransformer):
        def process(self, df):
            return

    o = _to_output_transformer(OT)
    w = FugueWorkflow()
    w.df([[0], [1]], "a:int", data_determiner=to_uuid).out_transform(
        o, params=dict(n=2)
    )
    assert_eq(
        """
    create [[0],[1]] schema a:int
    outtransform using OT(n=2)
    """,
        w,
    )

    w = FugueWorkflow()
    w.df([[0], [1]], "a:int", data_determiner=to_uuid).partition(
        by=["a"], presort="b DESC", num="ROWCOUNT/2"
    ).out_transform(mock_transformer, params=dict(n=2))
    assert_eq(
        """
    create [[0],[1]] schema a:int

    outtransform
        prepartition ROWCOUNT / 2 by a presort b desc
        using mock_transformer(n=2)
    """,
        w,
    )


def test_output():
    dag = FugueWorkflow()
    a = dag.create(mock_create1, params=dict(n=1))
    a.partition(num=4).output(mock_output)
    b = dag.create(mock_create1, params=dict(n=2))
    dag.output(a, b, using=mock_output, params=dict(n=3))
    assert_eq(
        """
    a=create using mock_create1(n=1)
    output prepartition 4 using mock_output
    output a, (create using mock_create1(n=2)) using mock_output(n=3)
    """,
        dag,
    )


def test_persist_checkpoint_broadcast():
    dag = FugueWorkflow()
    dag.create(mock_create1).persist()
    dag.create(mock_create1).weak_checkpoint(lazy=True, level="a.b")

    dag.create(mock_create1).broadcast()
    dag.create(mock_create1).weak_checkpoint(level="a.b").broadcast()

    dag.create(mock_create1).checkpoint()
    dag.create(mock_create1).strong_checkpoint(lazy=True)
    dag.create(mock_create1).strong_checkpoint(lazy=True, x="xy z")
    dag.create(mock_create1).strong_checkpoint(
        lazy=False, partition=PartitionSpec(num=5), single=True, x="xy z"
    ).broadcast()

    dag.create(mock_create1).deterministic_checkpoint()
    dag.create(mock_create1).deterministic_checkpoint(
        lazy=False, partition=PartitionSpec(num=4), single=True, namespace="n", x=2
    )
    assert_eq(
        """
    create using mock_create1 persist
    a=create using mock_create1 lazy persist (level="a.b")

    create using mock_create1 broadcast
    a=create using mock_create1 persist(level="a.b") broadcast

    create using mock_create1 checkpoint
    a= create using mock_create1 lazy strong checkpoint
    a=create using mock_create1 lazy checkpoint(x="xy z")
    a=create using mock_create1 checkpoint prepartition 5 single (x="xy z") broadcast

    create using mock_create1 deterministic checkpoint
    create using mock_create1 deterministic checkpoint "n"
        prepartition 4 single params x=2
    """,
        dag,
    )


def test_yield():
    dag = FugueWorkflow()
    dag.create(mock_create1).yield_dataframe_as("a", as_local=False)
    dag.create(mock_create1).yield_dataframe_as("aaa", as_local=True)
    dag.create(mock_create1).yield_file_as("aa")
    dag.create(mock_create1).yield_table_as("aaaa")
    dag.create(mock_create1).deterministic_checkpoint().yield_dataframe_as("c")
    dag.create(mock_create1).deterministic_checkpoint().yield_file_as("bb")
    dag.create(mock_create1).deterministic_checkpoint().yield_file_as("cc")

    assert_eq(
        """
    a=create using mock_create1 yield dataframe
    create using mock_create1 yield local dataframe as aaa
    create using mock_create1 yield file as aa
    create using mock_create1 yield table as aaaa
    c=create using mock_create1 deterministic checkpoint yield dataframe
    d=create using mock_create1 deterministic checkpoint yield file as bb
    create using mock_create1 deterministic checkpoint yield file as cc
    """,
        dag,
    )


def test_select_nested():
    dag = FugueWorkflow()
    a = dag.create(mock_create1, params=dict(n=1))
    b = dag.create(mock_create1, params=dict(n=2))
    dag.select("select * from (select * from a.b)")
    dag.select("select * from", dag.create(mock_create1), "AS bb")
    dag.select("select * from", dag.create(mock_create1), "TABLESAMPLE (5 PERCENT)")
    dag.select("select * from (select * from", dag.create(mock_create1), ")")
    assert_eq(
        """
    a=create using mock_create1(n=1)
    b=create using mock_create1(n=2)

    # nested query
    select * from (select * from a.b)
    select * from (create using mock_create1) AS bb
    select * from (create using mock_create1) TABLESAMPLE(5 PERCENT)
    select * from (select * from (create using mock_create1))
    """,
        dag,
    )


def test_select():
    dag = FugueWorkflow()
    a = dag.create(mock_create1, params=dict(n=1))
    b = dag.create(mock_create1, params=dict(n=2))
    dag.select("select * from a.b")
    dag.select("select * from a.b TABLESAMPLE (5 PERCENT) AS x")
    dag.select("select * from a.b AS x")
    dag.select("select * from", a, "AS a")  # fugue sql adds 'AS a'
    dag.select("select * from", a, "TABLESAMPLE (5 PERCENT) AS a")
    x = dag.select("select * from", a, "TABLESAMPLE (5 PERCENT) AS x")
    y = dag.select("select * FROM", x)
    z = dag.select("select * FROM", y, "where t = 100")
    dag.select("select a.* from", a, "AS a join", b, "AS b on a.a == b.a")

    dag.select("select * from", a, "AS a").persist().broadcast().show()
    dag.select("select * from", a, "AS a").weak_checkpoint(
        level="a.b.c"
    ).broadcast().show()
    assert_eq(
        """
    a=create using mock_create1(n=1)
    b=create using mock_create1(n=2)

    # assignment and table not found
    x=select * from a.b

    # sample and alias when table not found
    select * from a.b TABLESAMPLE (5 PERCENT) AS x
    select * from a.b AS x

    # when table is found
    select * from a
    select * from a TABLESAMPLE(5 PERCENT)
    select * from a TABLESAMPLE(5 PERCENT) AS x

    # no from
    select *
    select * where t=100

    # multiple dependencies
    select a.* from a join b on a.a==b.a

    # persist & checkpoint & broadcast
    select * from a persist broadcast print
    select * from a persist (level="a.b.c") broadcast print
    """,
        dag,
    )


def test_select_with():
    dag = FugueWorkflow()
    dag.select(
        "with x as ( select * from a ) , y as ( select * from b ) "
        "select * from x union select * from y"
    )
    assert_eq(
        """
    with
        x as (select * from a),
        y as (select * from b)
    select *   from x union select * from y

    """,
        dag,
    )


def test_select_plus_engine():
    class MockEngine(QPDPandasEngine):
        def __init__(self, execution_engine, p: int = 0):
            super().__init__(execution_engine)
            self.p = p

    register_sql_engine("_mock", lambda e, **kwargs: MockEngine(e, **kwargs))

    dag = FugueWorkflow()
    dag.select("select * from xyz", sql_engine=MockEngine).persist()
    dag.select("select * from xyz", sql_engine="_mock", sql_engine_params={"p": 2})
    dag.select("select * from xyz order by t limit 10", sql_engine=MockEngine)
    dag.select(
        "with a as ( select * from b ) select * from a order by t limit 10",
        sql_engine=MockEngine,
    )

    # temp = dag.select("select a , b from a", sql_engine=MockEngine)
    # temp.transform(mock_transformer2)

    # temp = dag.select("select aa , bb from a", sql_engine=MockEngine)
    # dag.select("select aa + bb as t from", temp)
    assert_eq(
        """
    connect MockEngine select * from xyz persist
    connect _mock(p=2) select * from xyz
    connect MockEngine select * from xyz order by t limit 10

    connect MockEngine with a as (select * from b) select * from a order by t limit 10

    # This is not supported
    # transform (connect MockEngine select a,b from a) using mock_transformer2

    # This is not supported
    # select aa+bb as t from (connect MockEngine select aa,bb from a)
    """,
        dag,
    )


def test_general_set_op():
    dag = FugueWorkflow()
    a = dag.create(mock_create1, params=dict(n=1))
    b = dag.create(mock_create1, params=dict(n=2))
    dag.select("select * from", a, "AS a union all select * from", b, "AS b")
    dag.select(
        "SELECT * FROM", dag.create(mock_create1), "union select * from", b, "AS b"
    )
    dag.select(
        "SELECT * FROM",
        dag.create(mock_create1),
        "intersect distinct SELECT * FROM",
        a.process(mock_processor1),
    )
    dag.select(
        "select * from",
        dag.create(mock_create1),
        "union SELECT * FROM",
        a.process(mock_processor1),
    )
    c = dag.create(mock_create1, params=dict(n=2))
    dag.select(
        "SELECT * FROM",
        c.transform(mock_transformer2),
        "union SELECT * FROM",
        c.process(mock_processor1),
    )
    assert_eq(
        """
    a=create using mock_create1(n=1)
    b=create using mock_create1(n=2)

    select * from a union all select * from b
    create using mock_create1 union select * from b
    create using mock_create1 intersect distinct process a using mock_processor1
    select * from (create using mock_create1) union process a using mock_processor1

    # operation on omitted dependencies should work as expected
    c=create using mock_create1(n=2)
    transform using mock_transformer2 union process using mock_processor1
    """,
        dag,
    )


def test_print():
    dag = FugueWorkflow()
    a = dag.create(mock_create1, params=dict(n=1))
    a.show()
    b = dag.create(mock_create1, params=dict(n=2))
    dag.show(a, b, n=5, with_count=True, title='"b   B')
    assert_eq(
        """
    a=create using mock_create1(n=1)
    print
    print 5 rows from a, (create using mock_create1(n=2)) rowcount title "\\"b   B"
    """,
        dag,
    )


def test_save():
    dag = FugueWorkflow()
    a = dag.create(mock_create1, params=dict(n=1))
    a.save("xx", fmt="parquet", mode="overwrite")
    a.save("xx", mode="append")
    a.save("xx", mode="error")
    a.save("xx.csv", fmt="csv", mode="error", single=True, header=True)
    a.partition(by=["x"]).save("xx", mode="overwrite")
    b = dag.create(mock_create1, params=dict(n=2)).save("xx", mode="overwrite")
    assert_eq(
        """
    a=create using mock_create1(n=1)
    save overwrite parquet "xx"
    save a append "xx"
    save a to "xx"
    save to single csv "xx.csv"(header=True)
    save prepartition by x overwrite "xx"
    save (create using mock_create1(n=2)) overwrite "xx"
    """,
        dag,
    )


def test_save_and_use():
    dag = FugueWorkflow()
    a = dag.create(mock_create1, params=dict(n=1))
    b = dag.create(mock_create1, params=dict(n=1))
    a = a.save_and_use("xx", fmt="parquet", mode="overwrite")
    b.save_and_use("xx", mode="append")
    b.save_and_use("xx", mode="error")
    a = a.save_and_use("xx.csv", fmt="csv", mode="error", single=True, header=True)
    a = a.partition(by=["x"]).save_and_use("xx", mode="overwrite")
    dag.create(mock_create1, params=dict(n=2)).save_and_use("xx", mode="overwrite")
    assert_eq(
        """
    a=create using mock_create1(n=1)
    b=create using mock_create1(n=1)
    a=save and use a overwrite parquet "xx"
    save and use b append "xx"
    save and use b to "xx"
    save and use a to single csv "xx.csv"(header=True)
    save and use prepartition by x overwrite "xx"
    save and use (create using mock_create1(n=2)) overwrite "xx"
    """,
        dag,
    )


def test_load():
    dag = FugueWorkflow()
    dag.load("xx")
    dag.load("xx", fmt="csv")
    dag.load("xx", columns="a:int,b:str")
    dag.load("xx", columns=["a", "b"], header=True)
    assert_eq(
        """
    load "xx"
    load csv "xx"
    load "xx" columns a:int, b:str
    load "xx"(header=True) columns a, b
    """,
        dag,
    )


def test_rename():
    dag = FugueWorkflow()
    a = dag.create(mock_create1)
    b = a.rename({"a": "aa", "b": "bb"})
    c = a.rename({"a": "aaa", "b": "bbb"})

    assert_eq(
        """
    a=create using mock_create1
    rename columns a:aa,b:bb
    rename columns a:aaa,b:bbb from a
    """,
        dag,
    )


def test_alter_columns():
    dag = FugueWorkflow()
    a = dag.create(mock_create1)
    a.alter_columns(Schema("a:str,b:str"))
    a.alter_columns(Schema("a:float,b:double"))

    assert_eq(
        """
    a=create using mock_create1
    alter columns a:str, b:str
    alter columns a:float, b:double from a
    """,
        dag,
    )


def test_drop():
    dag = FugueWorkflow()
    a = dag.create(mock_create1)
    b = a.drop(["a", "b"])
    c = a.drop(["a", "b"], if_exists=True)

    d = dag.create(mock_create1)
    e = d.dropna(how="any")
    f = d.dropna(how="all")
    g = d.dropna(how="any", subset=["a", "c"])
    assert_eq(
        """
    a=create using mock_create1
    drop columns a,b
    drop columns a,b if exists from a

    d=create using mock_create1
    drop rows if any null
    drop rows if all null from d
    drop rows if any nulls on a,c from d
    """,
        dag,
    )


def test_sample():
    dag = FugueWorkflow()
    a = dag.create(mock_create1)
    a.sample(frac=0.1, replace=False, seed=None)
    a.sample(n=5, replace=True, seed=7)

    assert_eq(
        """
    a=create using mock_create1
    sample 10 percent
    sample replace 5 rows seed 7 from a
    """,
        dag,
    )


def test_fill():
    dag = FugueWorkflow()
    a = dag.df([[None, 1], [1, None]], "a:int, b:int", data_determiner=to_uuid)
    b = a.fillna({"a": 99, "b": -99})
    assert_eq(
        """
    a=create [[NULL, 1],[1, NULL]] schema a:int, b:int
    fill nulls params a:99, b:-99 from a""",
        dag,
    )
    assert_eq(
        """
    create [[NULL, 1],[1, NULL]] schema a:int, b:int
    fill nulls (a:99, b:-99)""",
        dag,
    )


def test_head():
    dag = FugueWorkflow()
    a = dag.df(
        [[None, 1], [None, 2], [1, None], [1, 2]],
        "a:double, b:double",
        data_determiner=to_uuid,
    )
    b = a.partition(by=["a"], presort="b desc").take(1, na_position="first")
    c = b.take(1, presort="b desc", na_position="first")
    assert_eq(
        """
    a=create [[NULL, 1], [NULL, 2], [1, NULL], [1, 2]] schema a:double, b:double
    b=take 1 row from a prepartition by a presort b desc nulls first
    c=take 1 row from b presort b desc nulls first""",
        dag,
    )
    # anonymous
    assert_eq(
        """
    create [[NULL, 1], [NULL, 2], [1, NULL], [1, 2]] schema a:double, b:double
    take 1 row prepartition by a presort b desc nulls first
    take 1 row presort b desc nulls first""",
        dag,
    )


def test_module():
    # pylint: disable=no-value-for-parameter

    def create(wf: FugueWorkflow, n: int = 1) -> WorkflowDataFrame:
        return wf.df([[n]], "a:int")

    def merge(
        df1: WorkflowDataFrame, df2: WorkflowDataFrame, k: str = "aa"
    ) -> WorkflowDataFrames:
        return WorkflowDataFrames({k: df1, "bb": df2})

    def merge2(
        wf: FugueWorkflow, dfs: WorkflowDataFrames, k: int = 0
    ) -> WorkflowDataFrame:
        return dfs[k]

    def merge3(df1: WorkflowDataFrame, df2: WorkflowDataFrame) -> WorkflowDataFrames:
        return WorkflowDataFrames(df1, df2)

    @module()
    def out1(wf: FugueWorkflow, df: WorkflowDataFrame) -> None:
        df.show()

    dag = FugueWorkflow()
    a = create(dag)
    b = create(dag, n=2)
    dfs = merge(a, b, k="a1")
    dfs["a1"].show()
    dfs["bb"].show()
    df = merge2(dag, WorkflowDataFrames(a, b), k=1)
    out1(df)
    dfs = merge3(b, a)
    dfs[0].show()
    dfs[1].show()

    assert_eq(
        """
    a=sub using create
    b=sub using create(n=2)
    dfs=sub a,b using merge(k="a1")
    print dfs[a1]
    print dfs[bb]
    sub a,b using merge2(k=1)
    sub using out1
    dfs=sub df2:a,df1:b using merge3
    print dfs[0]
    print dfs[1]
    """,
        dag,
    )


def assert_eq(expr, expected: FugueWorkflow):
    global_vars, local_vars = get_caller_global_local_vars()
    sql = FugueSQLParser(expr, "fugueLanguage", ignore_case=True, parse_mode="auto")
    wf = FugueWorkflow()
    v = _Extensions(
        sql,
        FugueSQLHooks(),
        wf,
        dialect="spark",
        global_vars=global_vars,
        local_vars=local_vars,
    )
    obj = v.visit(sql.tree)
    assert expected.spec_uuid() == v.workflow.spec_uuid()


# schema: a:int,b:int
def mock_create1(n=2) -> List[List[Any]]:
    return [[n, n]]


def mock_create2(n=2) -> List[List[Any]]:
    return [[n]]


# schema: b:int
def mock_processor1(
    df1: List[List[Any]], df2: List[List[Any]], n=1
) -> Iterable[List[Any]]:
    for i in range(len(df1)):
        row1 = df1[i]
        row2 = df2[i]
        yield [max(row1[0], row2[0]) + n]


def mock_processor2(
    df1: List[List[Any]], df2: List[List[Any]], n=1
) -> Iterable[List[Any]]:
    for i in range(len(df1)):
        row1 = df1[i]
        row2 = df2[i]
        yield [max(row1[0], row2[0]) + n]


# schema: a:int
def mock_processor3(df: List[List[Any]]) -> List[List[Any]]:
    return df


def mock_transformer(df: LocalDataFrame, n=0) -> LocalDataFrame:
    pass


# schema: *
def mock_transformer2(df: LocalDataFrame, n=0) -> LocalDataFrame:
    pass


# schema: b:int
def mock_cotransformer1(
    df1: List[List[Any]], df2: List[List[Any]], n=1
) -> Iterable[List[Any]]:
    for i in range(len(df1)):
        row1 = df1[i]
        row2 = df2[i]
        yield [max(row1[0], row2[0]) + n]


def mock_cotransformer2(
    df1: List[List[Any]], df2: List[List[Any]], n=1
) -> Iterable[List[Any]]:
    for i in range(len(df1)):
        row1 = df1[i]
        row2 = df2[i]
        yield [max(row1[0], row2[0]) + n]


def mock_output(dfs: DataFrames, n=1) -> None:
    pass



from typing import Any, Iterable, List

import pandas as pd
from fugue.dataframe import ArrayDataFrame, DataFrame
from fugue.exceptions import FugueInterfacelessError
from fugue.execution import ExecutionEngine, NativeExecutionEngine
from fugue.extensions.creator import (
    Creator,
    _to_creator,
    creator,
    parse_creator,
    register_creator,
)
from pytest import raises
from triad.collections.dict import ParamDict
from triad.utils.hash import to_uuid


def test_creator():
    assert isinstance(t1, Creator)
    assert isinstance(t2, Creator)


def test_register():
    register_creator("x", T0)
    b = _to_creator("x")
    assert isinstance(b, Creator)

    register_creator("x", Creator, on_dup=ParamDict.IGNORE)

    raises(
        KeyError,
        lambda: register_creator("x", Creator, on_dup=ParamDict.THROW),
    )


def test_parse_creator():
    class _C(Creator):
        def __init__(self, x):
            self._x = x

        def create(self) -> DataFrame:
            raise NotImplementedError

        def __uuid__(self) -> str:
            return to_uuid(super().__uuid__(), self._x)

    @parse_creator.candidate(lambda x: isinstance(x, str) and x.startswith("(("))
    def _parse(obj):
        return _C(obj)

    a = _to_creator("((abc")
    b = _to_creator("((bc")
    c = _to_creator("((abc")

    assert isinstance(a, _C)
    assert isinstance(b, _C)
    assert isinstance(c, _C)
    assert to_uuid(a) == to_uuid(c)
    assert to_uuid(a) != to_uuid(b)


def test__to_creator():
    a = _to_creator(T0)
    assert isinstance(a, Creator)
    a = _to_creator(T0())

    assert isinstance(a, Creator)
    a = _to_creator(t1)
    assert isinstance(a, Creator)
    a._x = 1
    b = _to_creator(t1)
    assert isinstance(b, Creator)
    assert "_x" not in b.__dict__
    c = _to_creator(t1)
    assert isinstance(c, Creator)
    assert "_x" not in c.__dict__
    c._x = 1
    d = _to_creator("t1")
    assert isinstance(d, Creator)
    assert "_x" not in d.__dict__
    raises(FugueInterfacelessError, lambda: _to_creator("abc"))

    assert isinstance(_to_creator(t3), Creator)
    assert isinstance(_to_creator(t4), Creator)
    assert isinstance(_to_creator(t5), Creator)
    assert isinstance(_to_creator(t6), Creator)
    raises(FugueInterfacelessError, lambda: _to_creator(t6, "a:int"))
    assert isinstance(_to_creator(t7, "a:int"), Creator)
    raises(FugueInterfacelessError, lambda: _to_creator(t7))
    assert isinstance(_to_creator(t8), Creator)
    assert isinstance(_to_creator(t9), Creator)
    assert isinstance(_to_creator(t10), Creator)


def test_run_creator():
    o1 = _to_creator(t3)
    assert 4 == o1(4).as_array()[0][0]

    o1._params = ParamDict([("a", 2)], deep=False)
    o1._execution_engine = None
    assert 2 == o1.create().as_array()[0][0]

    o1 = _to_creator(t5)
    assert 4 == o1("dummy", 4)[0][0]
    o1._params = ParamDict([("a", 2)], deep=False)
    o1._execution_engine = NativeExecutionEngine()
    assert 2 == o1.create().as_array()[0][0]


def test__to_creator_determinism():
    a = _to_creator(t1, None)
    b = _to_creator(t1, None)
    c = _to_creator("t1", None)
    d = _to_creator("t2", None)
    assert a is not b
    assert to_uuid(a) == to_uuid(b)
    assert a is not c
    assert to_uuid(a) == to_uuid(c)
    assert to_uuid(a) != to_uuid(d)

    a = _to_creator(T0)
    b = _to_creator("T0")
    assert a is not b
    assert to_uuid(a) == to_uuid(b)


class T0(Creator):
    def create(self) -> DataFrame:
        return None


@creator("")
def t1() -> DataFrame:
    pass


@creator("a:int")
def t2(e: ExecutionEngine, a, b) -> List[List[Any]]:
    pass


def t3(a) -> DataFrame:
    return ArrayDataFrame([[a]], "a:int")


def t4(e: ExecutionEngine, a: int, b: str) -> DataFrame:
    pass


@creator("a:int")
def t5(e: ExecutionEngine, a) -> List[List[Any]]:
    assert e is not None
    return ArrayDataFrame([[a]], "a:int").as_array()


def t6() -> DataFrame:
    pass


def t7() -> List[List[Any]]:
    pass


# schema: a:int
def t8(e: ExecutionEngine, a, b) -> List[List[Any]]:
    pass


def t9(e: ExecutionEngine, a, b) -> Iterable[pd.DataFrame]:
    pass


def t10(e: ExecutionEngine, a, b, **kwargs) -> Iterable[pd.DataFrame]:
    pass



from typing import Any, Iterable, List, Optional, Union

import ibis
import pyarrow as pa
from triad import assert_or_throw

from fugue import (
    ArrowDataFrame,
    DataFrame,
    ExecutionEngine,
    NativeExecutionEngine,
    PartitionSpec,
    SQLEngine,
)
from fugue_ibis import IbisDataFrame, IbisExecutionEngine, IbisSQLEngine, IbisTable

from .dataframe import MockDuckDataFrame


class MockDuckSQLEngine(IbisSQLEngine):
    def __init__(self, execution_engine: ExecutionEngine) -> None:
        super().__init__(execution_engine)
        self._backend = ibis.duckdb.connect()

    @property
    def backend(self) -> ibis.BaseBackend:
        return self._backend

    @property
    def dialect(self) -> str:
        return "duckdb"

    @property
    def is_distributed(self) -> bool:
        return False

    def encode_column_name(self, name: str) -> str:
        return '"' + name.replace('"', '""') + '"'

    def to_df(self, df: Any, schema: Any = None) -> IbisDataFrame:
        if isinstance(df, MockDuckDataFrame):
            return df
        if isinstance(df, DataFrame):
            return self._register_df(
                df.as_arrow(), schema=schema if schema is not None else df.schema
            )
        if isinstance(df, pa.Table):
            return self._register_df(df, schema=schema)
        if isinstance(df, IbisTable):
            return MockDuckDataFrame(df, schema=schema)
        if isinstance(df, Iterable):
            adf = ArrowDataFrame(df, schema)
            return self._register_df(adf.native, schema=schema)
        raise NotImplementedError

    def persist(
        self,
        df: DataFrame,
        lazy: bool = False,
        **kwargs: Any,
    ) -> DataFrame:
        if isinstance(df, MockDuckDataFrame):
            res = ArrowDataFrame(df.as_arrow())
        else:
            res = self.to_df(df)
        res.reset_metadata(df.metadata)
        return res

    def sample(
        self,
        df: DataFrame,
        n: Optional[int] = None,
        frac: Optional[float] = None,
        replace: bool = False,
        seed: Optional[int] = None,
    ) -> DataFrame:
        assert_or_throw(
            (n is None and frac is not None and frac >= 0.0)
            or (frac is None and n is not None and n >= 0),
            ValueError(
                f"one and only one of n and frac should be non-negative, {n}, {frac}"
            ),
        )
        tn = self.get_temp_table_name()
        if frac is not None:
            sql = f"SELECT * FROM {tn} USING SAMPLE bernoulli({frac*100} PERCENT)"
        else:
            sql = f"SELECT * FROM {tn} USING SAMPLE reservoir({n} ROWS)"
        if seed is not None:
            sql += f" REPEATABLE ({seed})"
        idf = self.to_df(df)
        _res = f"WITH {tn} AS ({idf.native.compile()}) " + sql
        return self.to_df(self.backend.sql(_res))

    def _register_df(
        self, df: pa.Table, name: Optional[str] = None, schema: Any = None
    ) -> MockDuckDataFrame:
        tb = self.backend.register(df, name)
        return MockDuckDataFrame(tb)


class MockDuckExecutionEngine(IbisExecutionEngine):
    def __init__(self, conf: Any, force_is_ibis: bool = False):
        super().__init__(conf)

        self._force_is_ibis = force_is_ibis

    @property
    def is_distributed(self) -> bool:
        return False

    def create_non_ibis_execution_engine(self) -> ExecutionEngine:
        return NativeExecutionEngine(self.conf)

    def create_default_sql_engine(self) -> SQLEngine:
        return MockDuckSQLEngine(self)

    def is_non_ibis(self, ds: Any) -> bool:
        if self._force_is_ibis:
            return False
        return super().is_non_ibis(ds)

    def __repr__(self) -> str:
        return "MockDuckExecutionEngine"

    def load_df(
        self,
        path: Union[str, List[str]],
        format_hint: Any = None,
        columns: Any = None,
        **kwargs: Any,
    ) -> DataFrame:
        return self.non_ibis_engine.load_df(path, format_hint, columns, **kwargs)

    def save_df(
        self,
        df: DataFrame,
        path: str,
        format_hint: Any = None,
        mode: str = "overwrite",
        partition_spec: Optional[PartitionSpec] = None,
        force_single: bool = False,
        **kwargs: Any,
    ) -> None:
        partition_spec = partition_spec or PartitionSpec()
        _df = self._to_non_ibis_dataframe(df)
        return self.non_ibis_engine.save_df(
            _df, path, format_hint, mode, partition_spec, force_single, **kwargs
        )



import pytest

ibis = pytest.importorskip("ibis")
from fugue_dask import DaskExecutionEngine
from fugue_dask.ibis_engine import DaskIbisEngine
from fugue_ibis import IbisEngine
from fugue_test.ibis_suite import IbisTests


class DaskIbisTests(IbisTests.Tests):
    def make_engine(self):
        e = DaskExecutionEngine(conf=dict(test=True))
        return e

    def make_ibis_engine(self) -> IbisEngine:
        return DaskIbisEngine(self._engine)



from typing import Any

from fugue import ArrayBag, Bag
from fugue_test.bag_suite import BagTests


class ArrayBagTests(BagTests.Tests):
    def bg(self, data: Any = None) -> Bag:
        return ArrayBag(data)

    def test_array_bag_init(self):
        def _it():
            yield from [1, 2, 3]

        bg = self.bg([])
        assert bg.count() == 0
        assert bg.is_local
        assert bg.is_bounded
        assert bg.as_local() is bg
        assert bg.empty
        assert bg.native == []

        for x in [[1, 2, 3], _it(), set([1, 2, 3])]:
            bg = self.bg(x)
            assert bg.count() == 3
            assert bg.is_local
            assert bg.is_bounded
            assert bg.as_local() is bg
            assert not bg.empty
            assert 1 == bg.num_partitions
            assert isinstance(bg.native, list)

        bg = self.bg(x + 1 for x in [])
        assert bg.count() == 0
        bg = self.bg(x + 1 for x in [1, 2, 3])
        assert bg.count() == 3



# pylint: disable-all
from unittest import TestCase

import ibis
from fugue import ExecutionEngine, FugueWorkflow, register_default_sql_engine
from fugue_ibis import IbisEngine, as_fugue, as_ibis, run_ibis


class IbisTests(object):
    """Ibis test suite.
    Any new engine from :class:`~fugue_ibis.execution.ibis_engine.IbisEngine`
    should also pass this test suite.
    """

    class Tests(TestCase):
        @classmethod
        def setUpClass(cls):
            register_default_sql_engine(lambda engine: engine.sql_engine)
            cls._engine = cls.make_engine(cls)
            cls._ibis_engine = cls.make_ibis_engine(cls)

        @property
        def engine(self) -> ExecutionEngine:
            return self._engine  # type: ignore

        @property
        def ibis_engine(self) -> ExecutionEngine:
            return self._ibis_engine  # type: ignore

        @classmethod
        def tearDownClass(cls):
            cls._engine.stop()

        def make_engine(self) -> ExecutionEngine:  # pragma: no cover
            raise NotImplementedError

        def make_ibis_engine(self) -> IbisEngine:  # pragma: no cover
            raise NotImplementedError

        def test_run_ibis(self):
            def _test1(con: ibis.BaseBackend) -> ibis.Expr:
                tb = con.table("a")
                return tb

            def _test2(con: ibis.BaseBackend) -> ibis.Expr:
                tb = con.table("a")
                return tb.mutate(c=tb.a + tb.b)

            dag = FugueWorkflow()
            df = dag.df([[0, 1], [2, 3]], "a:long,b:long")
            res = run_ibis(_test1, ibis_engine=self.ibis_engine, a=df)
            res.assert_eq(df)
            df = dag.df([[0, 1], [2, 3]], "a:long,b:long")
            res = run_ibis(_test2, ibis_engine=self.ibis_engine, a=df)
            df2 = dag.df([[0, 1, 1], [2, 3, 5]], "a:long,b:long,c:long")
            res.assert_eq(df2)
            dag.run(self.engine)

        def test_run_as_ibis(self):
            dag = FugueWorkflow()
            df = dag.df([[0, 1], [2, 3]], "a:long,b:long")
            idf = as_ibis(df)
            res = as_fugue(idf)
            res.assert_eq(df)
            dag.run(self.engine)

            dag = FugueWorkflow()
            df1 = dag.df([[0, 1], [2, 3]], "a:long,b:long")
            df2 = dag.df([[0, ["x"]], [3, ["y"]]], "a:long,c:[str]")
            idf1 = as_ibis(df1)
            idf2 = as_ibis(df2)
            idf = idf1.inner_join(idf2, idf1.a == idf2.a)[idf1, idf2.c]
            res = as_fugue(idf)
            expected = dag.df([[0, 1, ["x"]]], "a:long,b:long,c:[str]")
            res.assert_eq(expected, check_order=True, check_schema=True)
            dag.run(self.engine)

            dag = FugueWorkflow()
            idf1 = dag.df([[0, 1], [2, 3]], "a:long,b:long").as_ibis()
            idf2 = dag.df([[0, ["x"]], [3, ["y"]]], "a:long,c:[str]").as_ibis()
            res = idf1.inner_join(idf2, idf1.a == idf2.a)[idf1, idf2.c].as_fugue()
            expected = dag.df([[0, 1, ["x"]]], "a:long,b:long,c:[str]")
            res.assert_eq(expected, check_order=True, check_schema=True)
            dag.run(self.engine)

        def test_literal(self):
            dag = FugueWorkflow()
            idf1 = dag.df([[0, 1], [2, 3]], "a:long,b:long").as_ibis()
            res = idf1.mutate(c=idf1.b + 10).as_fugue()
            expected = dag.df([[0, 1, 11], [2, 3, 13]], "a:long,b:long,c:long")
            res.assert_eq(expected, check_order=True, check_schema=True)
            dag.run(self.engine)



import pickle

import duckdb
import pandas as pd
import pyarrow as pa
from pytest import raises

import fugue.api as fa
from fugue import ArrowDataFrame, DataFrame, FugueWorkflow, fsql
from fugue.api import engine_context
from fugue.plugins import infer_execution_engine
from fugue_duckdb import DuckExecutionEngine
from fugue_duckdb.dataframe import DuckDataFrame
from fugue_test.builtin_suite import BuiltInTests
from fugue_test.execution_suite import ExecutionEngineTests


class DuckExecutionEngineTests(ExecutionEngineTests.Tests):
    @classmethod
    def setUpClass(cls):
        cls._con = duckdb.connect()
        cls._engine = cls.make_engine(cls)
        fa.set_global_engine(cls._engine)

    @classmethod
    def tearDownClass(cls):
        fa.clear_global_engine()
        cls._con.close()

    def make_engine(self):
        e = DuckExecutionEngine(
            {
                "test": True,
                "fugue.duckdb.pragma.threads": 2,
                "fugue.duckdb.extensions": ", json , , httpfs",
            },
            self._con,
        )
        return e

    def test_properties(self):
        assert not self.engine.is_distributed
        assert not self.engine.map_engine.is_distributed
        assert not self.engine.sql_engine.is_distributed

    def test_duckdb_extensions(self):
        df = fa.fugue_sql(
            """
        SELECT COUNT(*) AS ct FROM duckdb_extensions()
        WHERE loaded AND extension_name IN ('httpfs', 'json')
        """,
            as_fugue=True,
        )
        assert 2 == df.as_array()[0][0]

    def test_duck_to_df(self):
        e = self.engine
        a = e.to_df([[1, 2, 3]], "a:double,b:double,c:int")
        assert isinstance(a, DuckDataFrame)
        b = e.to_df(a.native_as_df())
        assert isinstance(b, DuckDataFrame)

    def test_table_operations(self):
        se = self.engine.sql_engine
        df = fa.as_fugue_df([[0, 1]], schema="a:int,b:long")
        assert se._get_table("_t_x") is None
        se.save_table(df, "_t_x")
        assert se._get_table("_t_x") is not None
        res = se.load_table("_t_x").as_array()
        assert [[0, 1]] == res
        df = fa.as_fugue_df([[1, 2]], schema="a:int,b:long")
        se.save_table(df, "_t_x")
        res = se.load_table("_t_x").as_array()
        assert [[1, 2]] == res
        with raises(Exception):
            se.save_table(df, "_t_x", mode="error")

    def test_intersect_all(self):
        e = self.engine
        a = e.to_df([[1, 2, 3], [4, None, 6], [4, None, 6]], "a:double,b:double,c:int")
        b = e.to_df(
            [[1, 2, 33], [4, None, 6], [4, None, 6], [4, None, 6]],
            "a:double,b:double,c:int",
        )
        raises(NotImplementedError, lambda: e.intersect(a, b, distinct=False))
        # DuckDB 0.5.0 stopped support INTERSECT ALL
        # c = e.intersect(a, b, distinct=False)
        # df_eq(
        #     c,
        #     [[4, None, 6], [4, None, 6]],
        #     "a:double,b:double,c:int",
        #     throw=True,
        # )


class DuckBuiltInTests(BuiltInTests.Tests):
    @classmethod
    def setUpClass(cls):
        cls._con = duckdb.connect()
        cls._engine = cls.make_engine(cls)

    @classmethod
    def tearDownClass(cls):
        cls._con.close()

    def make_engine(self):
        e = DuckExecutionEngine(
            {"test": True, "fugue.duckdb.pragma.threads": 2}, self._con
        )
        return e

    def test_special_types(self):
        def assert_data(df: DataFrame) -> None:
            assert df.schema == "a:datetime,b:bytes,c:[long]"

        df = pd.DataFrame(
            [["2020-01-01", pickle.dumps([1, 2]), [1, 2]]], columns=list("abc")
        )
        df["a"] = pd.to_datetime(df.a)

        with FugueWorkflow() as dag:
            x = dag.df(df)
            result = dag.select("SELECT * FROM ", x)
            result.output(assert_data)
        dag.run(self.engine)


def test_builtin_connection():
    dag = FugueWorkflow()
    df = dag.df([[0], [1]], "a:long")
    df = dag.select("SELECT * FROM", df, "WHERE a<1")
    df.assert_eq(dag.df([[0]], "a:long"))

    dag.run("duckdb")
    dag.run(("native", "duck"))


def test_configs():
    dag = FugueWorkflow()
    df = dag.df([[None], [1]], "a:double")
    df = dag.select("SELECT * FROM ", df, "ORDER BY a LIMIT 1")
    df.assert_eq(dag.df([[None]], "a:double"))

    dag.run(
        "duckdb",
        {
            "fugue.duckdb.pragma.threads": 2,
            "fugue.duckdb.pragma.default_null_order": "NULLS FIRST",
        },
    )

    dag = FugueWorkflow()
    df = dag.df([[None], [1]], "a:double")
    df = dag.select("SELECT * FROM ", df, "ORDER BY a LIMIT 1")
    df.assert_eq(dag.df([[1]], "a:double"))

    dag.run(
        "duckdb",
        {
            "fugue.duckdb.pragma.threads": 2,
            "fugue.duckdb.pragma.default_null_order": "NULLS LAST",
        },
    )

    with raises(ValueError):
        # invalid config format
        dag.run("duckdb", {"fugue.duckdb.pragma.threads;xx": 2})

    with raises(Exception):
        # non-existent config
        dag.run("duckdb", {"fugue.duckdb.pragma.threads_xx": 2})


def test_annotations():
    con = duckdb.connect()

    def cr(c: duckdb.DuckDBPyConnection) -> duckdb.DuckDBPyRelation:
        return c.query("SELECT 1 AS a")

    def pr(df: duckdb.DuckDBPyRelation) -> pa.Table:
        return df.arrow()

    def ot(df: duckdb.DuckDBPyRelation) -> None:
        assert 1 == df.df().shape[0]

    dag = FugueWorkflow()
    dag.create(cr).process(pr).output(ot)

    dag.run(con)


def test_output_types():
    pdf = pd.DataFrame([[0]], columns=["a"])
    con = duckdb.connect()

    res = fsql(
        """
    CREATE [[0]] SCHEMA a:int
    YIELD DATAFRAME AS a
    CREATE [[0]] SCHEMA b:int
    YIELD LOCAL DATAFRAME AS b
    """
    ).run(con)

    assert isinstance(res["a"], DuckDataFrame)
    assert isinstance(res["b"], ArrowDataFrame)

    x = fa.union(pdf, pdf, engine=con)
    assert isinstance(x, duckdb.DuckDBPyRelation)

    con.close()

    # ephemeral
    res = fsql(
        """
    CREATE [[0]] SCHEMA a:int
    YIELD DATAFRAME AS a
    CREATE [[0]] SCHEMA b:int
    YIELD LOCAL DATAFRAME AS b
    """
    ).run("duck")

    assert isinstance(res["a"], ArrowDataFrame)
    assert isinstance(res["b"], ArrowDataFrame)

    x = fa.union(pdf, pdf, engine="duckdb")
    assert isinstance(x, pa.Table)

    # in context
    with engine_context("duck"):
        res = fsql(
            """
        CREATE [[0]] SCHEMA a:int
        YIELD DATAFRAME AS a
        CREATE [[0]] SCHEMA b:int
        YIELD LOCAL DATAFRAME AS b
        """
        ).run()

        assert isinstance(res["a"], DuckDataFrame)
        assert isinstance(res["b"], ArrowDataFrame)

        x = fa.union(pdf, pdf)
        assert isinstance(x, duckdb.DuckDBPyRelation)


def test_infer_engine():
    con = duckdb.connect()
    df = con.from_df(pd.DataFrame([[0]], columns=["a"]))
    assert infer_execution_engine([df]) == "duckdb"

    fdf = DuckDataFrame(df)
    assert infer_execution_engine([fdf]) == "duckdb"

    con.close()



from typing import Any, Dict, Iterable

from fugue.dataframe import ArrayDataFrame, DataFrame, DataFrames
from fugue.exceptions import FugueInterfacelessError
from fugue.execution import ExecutionEngine, NativeExecutionEngine
from fugue.extensions.outputter import (
    Outputter,
    _to_outputter,
    outputter,
    parse_outputter,
    register_outputter,
)
from pytest import raises
from triad.collections.dict import ParamDict
from triad.utils.hash import to_uuid


def test_outputter():
    assert isinstance(t1, Outputter)
    assert isinstance(t2, Outputter)


def test_register():
    register_outputter("x", MockOutputter)
    b = _to_outputter("x")
    assert isinstance(b, MockOutputter)

    raises(
        KeyError,
        lambda: register_outputter("x", MockOutputter, on_dup=ParamDict.THROW),
    )


def test__to_outputter():
    a = _to_outputter(MockOutputter)
    assert isinstance(a, MockOutputter)
    b = _to_outputter("MockOutputter")
    assert isinstance(b, MockOutputter)

    a = _to_outputter(T0)
    assert isinstance(a, Outputter)
    a = _to_outputter(T0())

    assert isinstance(a, Outputter)
    a = _to_outputter(t1)
    assert isinstance(a, Outputter)
    a._x = 1
    b = _to_outputter(t1)
    assert isinstance(b, Outputter)
    assert "_x" not in b.__dict__
    c = _to_outputter(t1)
    assert isinstance(c, Outputter)
    assert "_x" not in c.__dict__
    c._x = 1
    d = _to_outputter("t1")
    assert isinstance(d, Outputter)
    assert "_x" not in d.__dict__
    raises(FugueInterfacelessError, lambda: _to_outputter("abc"))

    assert isinstance(_to_outputter(t3), Outputter)
    assert isinstance(_to_outputter(t4), Outputter)
    assert isinstance(_to_outputter(t5), Outputter)
    assert isinstance(_to_outputter(t6), Outputter)
    assert isinstance(_to_outputter(t7), Outputter)


def test_run_outputter():
    df = ArrayDataFrame([[0]], "a:int")
    dfs = DataFrames(df1=df, df2=df)
    dfs2 = DataFrames(df, df)
    assert not dfs2.has_key

    class Ct(object):
        pass

    c = Ct()
    o1 = _to_outputter(t3)
    o1(df, df, 2, c)
    assert 4 == c.value
    c.value = 0
    o1._params = ParamDict([("a", 2), ("b", c)], deep=False)
    o1._execution_engine = None
    o1.process(dfs)
    assert 4 == c.value
    c.value = 0
    o1._params = ParamDict([("a", 2), ("b", c)], deep=False)
    o1._execution_engine = None
    o1.process(dfs2)
    assert 4 == c.value

    c = Ct()
    o1 = _to_outputter(t5)
    o1("dummy", dfs, 2, c)
    assert 4 == c.value
    c.value = 0
    o1("dummy", dfs2, 2, c)
    assert 4 == c.value
    c.value = 0
    o1._params = ParamDict([("a", 2), ("b", c)], deep=False)
    o1._execution_engine = NativeExecutionEngine()
    o1.process(dfs)
    assert 4 == c.value
    c.value = 0
    o1._params = ParamDict([("a", 2), ("b", c)], deep=False)
    o1._execution_engine = NativeExecutionEngine()
    o1.process(dfs2)
    assert 4 == c.value


def test__to_outputter_determinism():
    a = _to_outputter(t1)
    b = _to_outputter(t1)
    c = _to_outputter("t1")
    d = _to_outputter("t2")
    assert a is not b
    assert to_uuid(a) == to_uuid(b)
    assert a is not c
    assert to_uuid(a) == to_uuid(c)
    assert to_uuid(a) != to_uuid(d)

    a = _to_outputter(MockOutputter)
    b = _to_outputter("MockOutputter")
    assert a is not b
    assert to_uuid(a) == to_uuid(b)


def test_to_outputter_validation():
    @outputter(input_has=" a , b ")
    def ov1(df: Iterable[Dict[str, Any]]) -> None:
        pass

    # input_has: a , b
    def ov2(df: Iterable[Dict[str, Any]]) -> None:
        pass

    class MockOutputterV(Outputter):
        @property
        def validation_rules(self):
            return {"input_is": "a:int,b:int"}

        def process(self, dfs):
            pass

    a = _to_outputter(ov1, None)
    assert {"input_has": ["a", "b"]} == a.validation_rules
    b = _to_outputter(ov2, None)
    assert {"input_has": ["a", "b"]} == b.validation_rules
    c = _to_outputter(MockOutputterV)
    assert {"input_is": "a:int,b:int"} == c.validation_rules


def test_parse_outputter():
    class _C(Outputter):
        def __init__(self, x):
            self._x = x

        def process(self, dfs) -> DataFrame:
            raise NotImplementedError

        def __uuid__(self) -> str:
            return to_uuid(super().__uuid__(), self._x)

    @parse_outputter.candidate(lambda x: isinstance(x, str) and x.startswith("(("))
    def _parse(obj):
        return _C(obj)

    a = _to_outputter("((abc")
    b = _to_outputter("((bc")
    c = _to_outputter("((abc")

    assert isinstance(a, _C)
    assert isinstance(b, _C)
    assert isinstance(c, _C)
    assert to_uuid(a) == to_uuid(c)
    assert to_uuid(a) != to_uuid(b)


class T0(Outputter):
    def process(self, dfs):
        pass


@outputter()
def t1(df: Iterable[Dict[str, Any]]) -> None:
    pass


@outputter()
def t2(e: ExecutionEngine, df1: DataFrame, df2: DataFrame) -> None:
    pass


def t3(df1: DataFrame, df2: DataFrame, a, b) -> None:
    b.value = df1.count() + df2.count() + a


def t4(e: ExecutionEngine, df1: DataFrame, df2: DataFrame, a: int, b: str) -> None:
    pass


def t5(e: ExecutionEngine, dfs: DataFrames, a, b) -> None:
    assert e is not None
    b.value = sum(x.count() for x in dfs.values()) + a


def t6(dfs: DataFrames) -> None:
    pass


def t7(df7: DataFrames, **kwargs) -> None:
    pass


class MockOutputter(Outputter):
    def process(self, dfs):
        pass
